/*
 * I2C readout for MPPT Charger Proto 1 PCB with Linksprite LCD display
 * 
 * For Arduino Leonardo
 * 
 * IO Assignment
 *   D0 - NIGHT Sense
 *   D1 - ALERT_N Sense
 *   D2 - Charger SDA
 *   D3 - Charger SCL
 *   D4 - LCD Display SD card CS (unused)
 *   D5 - LCD Display LCD CS
 *   D6 - LCD Display TFT_D/C pin
 *   D7 - LCD Display Backlight enable
 *   D9 - +5V Sense
 *   D10 - LCD Display SPI CS
 *   D11 - LCD Display SPI MOSI
 *   D12 - LCD Display SPI MISO
 *   D13 - LCD Display SPI SCK
 *   A0 - LCD Display Touch Y-
 *   A1 - LCD Display Touch X-
 *   A2 - LCD Display Touch Y+
 *   A3 - LCD Display Touch X+
 *   A4 -
 *   A5 -  
 * 
 * See the "DisplayHelp" subroutine at the end for the command interface list.
 * 
 * Display Organization
 *   VS : XXXXX
 *   IS : XXXXX
 *   VB : XXXXX
 *   IB : XXXXX
 *   IC : XXXXX
 *   CHG: N
 *   BCK: NNN 1 2
 *   TI : NNN
 *   TE : NNN
 *   VM : XXXXX
 *   TH : XXXXX
 *   ST : HHHH
 *   WD : NNN  (if Watchdog Enable is set)
 *   IO : 5 A N
 *   "WDT Reset"  if system watchdog triggered flag detected
 *   "WDT Power"  if power watchdog triggered flag detected
 */
#include <TFTv2.h>
#include <SPI.h>
#include <TouchScreen.h>
#include <Wire.h>


#define ALERT_SNS 1
#define NIGHT_SNS 0
#define PWR_SNS   9


#define I2C_ADDR 0x12


// RO state 16-bit register indices (index into local array)
#define SMB_INDEX_ID       0
#define SMB_INDEX_STATUS   1
#define SMB_INDEX_BUCK     2
#define SMB_INDEX_VS       3
#define SMB_INDEX_IS       4
#define SMB_INDEX_VB       5
#define SMB_INDEX_IB       6
#define SMB_INDEX_IC       7
#define SMB_INDEX_I_TEMP   8
#define SMB_INDEX_E_TEMP   9
#define SMB_INDEX_VM       10
#define SMB_INDEX_TH       11


// RW PARAM 16-bit register SMBus address (must match PARAM order)
#define SMB_ADDR_BULK_V    24
#define SMB_ADDR_FLT_V     26
#define SMB_ADDR_POFF_V    28
#define SMB_ADDR_PON_V     30
// RW Watchdog 8-bit register SMBus address
#define SMB_ADDR_WD_EN     33
#define SMB_ADDR_WD_TO     35
#define SMB_ADDR_WD_PWROFF 36

#define SMB_RO_NUM          12


// Status register masks
#define SMB_ST_SWD_DET_MASK  0x8000
#define SMB_ST_PWD_TRIG_MASK 0x4000
#define SMB_ST_BAD_BATT_MASK 0x2000
#define SMB_ST_ET_MISS_MASK  0x1000
#define SMB_ST_WD_RUN_MASK   0x0100
#define SMB_ST_PWR_EN_MASK   0x0080
#define SMB_ST_ALERT_MASK    0x0040
#define SMB_ST_PCTRL_MASK    0x0020
#define SMB_ST_T_LIM_MASK    0x0010
#define SMB_ST_NIGHT_MASK    0x0008
#define SMB_ST_CHG_ST_MASK   0x0007


// Display constants
#define CHAR_SIZE_MULT   2
#define CHAR_X_W_OFFSET  5
#define CHAR_Y_OFFSET    10
#define CHAR_X_W         6
#define CHAR_Y_W         8
#define LINE_SPACING     9
#define CHAR_DATA_OFF    5*CHAR_X_W


// Command processing
#define CMD_ST_IDLE 0
#define CMD_ST_CMD 1
#define CMD_ST_VAL 2

#define TERM_CHAR 0x0D


#define SMB_NUM_LABELS   14
#define LB_WD_COUNT_LINE 12
#define LB_IO_PIN_LINE   13
#define LB_SYS_WD_LINE   14
#define LB_PWR_WD_LINE   15
char *labelsP[SMB_NUM_LABELS] = {
  "VS :",
  "IS :",
  "VB :",
  "IB :",
  "IC :",
  "CHG:",
  "BCK:",
  "TI :",
  "TE :",
  "VM :",
  "TH :",
  "ST :",
  "WD :",
  "IO :"
};

// operating modes
#define MODE_LCD       0
#define MODE_WAIT_TAP  1
#define MODE_SER       2
#define MODE_IDLE      3

// Serial logging range - contiguous numbers somewhere in the RO register array
#define SER_LOG_MIN SMB_INDEX_STATUS
#define SER_LOG_MAX SMB_INDEX_IC

uint16_t data[SMB_RO_NUM];
uint16_t prevData[SMB_RO_NUM];

// Watchdog logging state
bool watchdogEnabled;
uint8_t watchdogCount;

// IO Pin state
bool alertAsserted;
bool nightAsserted;
bool pwrEnabled;

// Command processor
int CmdState = CMD_ST_IDLE;
byte CurCmd;
int CmdArg;
uint16_t CmdVal;
boolean CmdHasVal;
char CmdBuf[32];

// Mode
int logMode = MODE_LCD;
int logCount;
bool logTapTrigEn = false;

unsigned long prevT;

//
// Objects
//
TouchScreen ts = TouchScreen(XP, YP, XM, YM); //init TouchScreen port pins


//
// Arduino Entry points
//
void setup()
{
  Serial.begin(115200);
  Wire.begin();

  // Initialize pin state to the opposite of current value to force initial lcd update
  pinMode(ALERT_SNS, INPUT);
  pinMode(NIGHT_SNS, INPUT);
  pinMode(PWR_SNS, INPUT);
  alertAsserted = !(digitalRead(ALERT_SNS) == LOW);
  nightAsserted = !(digitalRead(NIGHT_SNS) == HIGH);
  pwrEnabled    = !(digitalRead(PWR_SNS) == HIGH);
  
  TFT_BL_ON;
  Tft.TFTinit();

  PrintLabels();

  // Time for reprogramming if I2C hangs in main loop
  delay(5000);

  // Initialize our previous data array with the inverted values of an initial reading
  (void) ReadArrayBurst();
  for (int i=0; i<SMB_RO_NUM; i++) {
    prevData[i] = ~data[i];
  }

  prevT = millis();
  
  Serial.println(F("Starting..."));
}

void loop()
{

  EvalCommand();

  switch (logMode) {
    case MODE_LCD:
      EvalLcdLogging();
      break;
    case MODE_WAIT_TAP:
      EvalWaitTap();
      break;
    case MODE_SER:
      EvalSerialLogging();
      break;
  }
}


//
// Subroutines
//
void EvalLcdLogging()
{
  PrintIOPins(LB_IO_PIN_LINE);

  if (InactivityTimeout(prevT, 100)) {
    prevT = millis();
    if (ReadArrayBurst()) {
      PrintData();
    
      // Watchdog related
      if (data[SMB_INDEX_STATUS] & SMB_ST_SWD_DET_MASK) {
        PrintWdtResetDetected(true);
      }
      if (data[SMB_INDEX_STATUS] & SMB_ST_PWD_TRIG_MASK) {
        PrintWatchdogDetected(true);
      }
      if (data[SMB_INDEX_STATUS] & SMB_ST_WD_RUN_MASK) {
        PrintWatchdogCount(true);
      } else {
        PrintWatchdogCount(false);
      }
    }
  }
}


void EvalWaitTap()
{
  Point p = ts.getPoint();

  if (p.z > __PRESURE) {
    logMode = MODE_SER;
    SetupDisplaySerLog();
  }
}


void EvalSerialLogging()
{
  int i;
  
  if (ReadSerLogArrayBurst()) {
    (void) Read16(SMB_INDEX_VM*2, &data[SMB_INDEX_VM]);
    for (int i=SER_LOG_MIN; i<=SER_LOG_MAX; i++) {
      if (i == SMB_INDEX_BUCK) {
        // Only display pwm value
        Serial.print(data[i] >> 6);
      } else {
        Serial.print((int16_t) data[i]);
      }
 //     if (i != SER_LOG_MAX) {
        Serial.print(F(","));
 //     }
    }
    Serial.print(data[SMB_INDEX_VM]);
    Serial.println();

    if (--logCount == 0) {
      logMode = MODE_LCD;
      SetupDisplayLcdLog();
    } else {
      if (logCount % 10 == 0) {
        PrintNum16(0, 0, logCount);
      }
    }
  }
}


boolean InactivityTimeout(unsigned long prevT, unsigned long timeout) {
  unsigned long curT = millis();
  unsigned long deltaT;
  
  if (curT > prevT) {
    deltaT = curT - prevT;
  } else {
    // Handle wrap
    deltaT = ~(prevT - curT) + 1;
  }
  
  return (deltaT >= timeout);
}


bool Read16(uint8_t index, uint16_t* val)
{
  uint8_t u, l;
  
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(index);
  Wire.endTransmission();

  if (Wire.requestFrom(I2C_ADDR, 2) != 2) {
    *val = 0;
    return(false);
  } else {
    u = Wire.read();
    l = Wire.read();
    *val = u << 8 | l;
    return(true);
  }
}


bool Read8(uint8_t index, uint8_t* val)
{
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(index);
  Wire.endTransmission();

  if (Wire.requestFrom(I2C_ADDR, 1) != 1) {
    *val = 0;
    return(false);
  } else {
    *val = Wire.read();
    return(true);
  }
}


bool Write16(uint8_t index, uint16_t val)
{
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(index);
  Wire.write(val >> 8);
  Wire.write(val & 0xFF);
  return(Wire.endTransmission() == 0);
}


bool Write8(uint8_t index, uint8_t val)
{
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(index);
  Wire.write(val);
  return(Wire.endTransmission() == 0);
}


bool ReadArraySingle()
{
  bool retVal = true;
  uint8_t i;

  for (i=0; i<SMB_RO_NUM; i++) {
    if (!Read16(2*i, &data[i])) {
      retVal = false;
      break;
    }
  }

  return(retVal);
}


bool ReadArrayBurst()
{
  uint8_t i;
  uint8_t u, l;
  
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(0);
  Wire.endTransmission();

  if (Wire.requestFrom(I2C_ADDR, SMB_RO_NUM*2) != SMB_RO_NUM*2) {
    return(false);
  } else {
    for (i=0; i<SMB_RO_NUM; i++) {
      u = Wire.read();
      l = Wire.read();
      data[i] = u << 8 | l;
    }
    return(true);
  }
}


bool ReadSerLogArrayBurst()
{
  uint8_t i;
  uint8_t n = SER_LOG_MAX - SER_LOG_MIN + 1;
  uint8_t u, l;
  
  Wire.beginTransmission(I2C_ADDR);
  Wire.write(SER_LOG_MIN*2);
  Wire.endTransmission();

  if (Wire.requestFrom(I2C_ADDR, n*2) != n*2) {
    return(false);
  } else {
    for (i=SER_LOG_MIN; i<=SER_LOG_MAX; i++) {
      u = Wire.read();
      l = Wire.read();
      data[i] = u << 8 | l;
    }
    return(true);
  }
}


void SetupDisplayLcdLog()
{
  ClearScreen();
  PrintLabels();

  // Initialize our previous data array with the inverted values of data to force lcd update
  for (int i=0; i<SMB_RO_NUM; i++) {
    prevData[i] = ~data[i];
  }

  // Flip the IO status to force update
  alertAsserted = !alertAsserted;
  nightAsserted = !nightAsserted;
  pwrEnabled    = !pwrEnabled;
}


void SetupDisplayWaitTap()
{
  ClearScreen();
  Tft.drawString("Tap to log", CHAR_X_W_OFFSET, CHAR_Y_OFFSET, CHAR_SIZE_MULT, WHITE);
}


void SetupDisplaySerLog()
{
  ClearScreen();
  Tft.drawString("LOG:", CHAR_X_W_OFFSET, CHAR_Y_OFFSET, CHAR_SIZE_MULT, WHITE);
}


void ClearScreen()
{
  Tft.fillRectangle(0, 0, 240, 320, BLACK);
}


void PrintLabels()
{
  uint8_t i;
  uint16_t x, y;

  x = CHAR_X_W_OFFSET;
  y = CHAR_Y_OFFSET;
  for (i=0; i<SMB_NUM_LABELS; i++) {
    Tft.drawString(labelsP[i], x, y, CHAR_SIZE_MULT, WHITE);
    y += LINE_SPACING*CHAR_SIZE_MULT;
  }
}


void PrintData()
{
  PrintConditionalNum16(0, SMB_INDEX_VS);
  PrintConditionalNum16(1, SMB_INDEX_IS);
  PrintConditionalNum16(2, SMB_INDEX_VB);
  PrintConditionalNum16(3, SMB_INDEX_IB);
  PrintConditionalNum16(4, SMB_INDEX_IC);
  if (data[SMB_INDEX_STATUS] != prevData[SMB_INDEX_STATUS]) {
    prevData[SMB_INDEX_STATUS] = data[SMB_INDEX_STATUS];
    PrintChargeState(5, data[SMB_INDEX_STATUS] & SMB_ST_CHG_ST_MASK);
    PrintHex16(11, data[SMB_INDEX_STATUS]);
  }
  if (data[SMB_INDEX_BUCK] != prevData[SMB_INDEX_BUCK]) {
    prevData[SMB_INDEX_BUCK] = data[SMB_INDEX_BUCK];
    PrintBuckStatus(6, data[SMB_INDEX_BUCK]);
  }
  PrintConditionalNum16(7, SMB_INDEX_I_TEMP);
  PrintConditionalNum16(8, SMB_INDEX_E_TEMP);
  PrintConditionalNum16(9, SMB_INDEX_VM);
  PrintConditionalNum16(10, SMB_INDEX_TH);
}


void PrintConditionalNum16(uint8_t l, int index)
{
  if (data[index] != prevData[index]) {
    prevData[index] = data[index];
    PrintNum16(l, 0, data[index]);
  }
}


void PrintNum16(uint8_t l, uint8_t col, uint16_t v)
{
  uint16_t x, y;
  char numString[6];

  x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT + col*CHAR_X_W*CHAR_SIZE_MULT;
  y = CHAR_Y_OFFSET + (l * LINE_SPACING*CHAR_SIZE_MULT);
  (void) itoa(v, numString, 10);
  Tft.fillRectangle(x, y, 6*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  Tft.drawString(numString, x, y, CHAR_SIZE_MULT, WHITE);
}


void PrintHex16(uint8_t l, uint16_t v)
{
  bool leadingZero = true;
  uint8_t n;
  uint16_t x, y;
  char numString[5];

  x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT;
  y = CHAR_Y_OFFSET + (l * LINE_SPACING*CHAR_SIZE_MULT);
  numString[4] = 0; // term

  n = (v >> 12) & 0x0F;
  if (n != 0) {
    numString[0] = digitToHexChar(n);
    leadingZero = false;
  } else {
    numString[0] = ' ';
  }

  n = (v >> 8) & 0x0F;
  if ((n != 0) || (!leadingZero)) {
    numString[1] = digitToHexChar(n);
    leadingZero = false;
  } else {
    numString[1] = ' ';
  }

  n = (v >> 4) & 0x0F;
  if ((n != 0) || (!leadingZero)) {
    numString[2] = digitToHexChar(n);
  } else {
    numString[2] = ' ';
  }

  n = v & 0x0F;  
  numString[3] = digitToHexChar(n);

  Tft.fillRectangle(x, y, 5*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  Tft.drawString(numString, x, y, CHAR_SIZE_MULT, WHITE);
}


char digitToHexChar(uint8_t n)
{
  n = n & 0x0F;
  
  if (n < 10) {
    return('0' + n);
  } else {
    return('A' + (n - 10));
  }
}


void PrintChargeState(uint8_t l, uint8_t s)
{
  uint16_t x, y;

  x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT;
  y = CHAR_Y_OFFSET + (l * LINE_SPACING*CHAR_SIZE_MULT);

  Tft.fillRectangle(x, y, 6*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  
  switch (s) {
    case 0:
      Tft.drawString("NIGHT", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 1:
      Tft.drawString("IDLE", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 2:
      Tft.drawString("VSRCV", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 3:
      Tft.drawString("SCAN", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 4:
      Tft.drawString("BULK", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 5:
      Tft.drawString("ABS", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    case 6:
      Tft.drawString("FLOAT", x, y, CHAR_SIZE_MULT, WHITE);
      break;
    default:
      Tft.drawString("?????", x, y, CHAR_SIZE_MULT, WHITE);
      break;
  }
}


void PrintBuckStatus(uint8_t l, uint16_t v)
{
  uint16_t x, y;

  // Print PWM
  PrintNum16(l, 0, (v >> 6));

  x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT + 6*CHAR_X_W*CHAR_SIZE_MULT;
  y = CHAR_Y_OFFSET + (l * LINE_SPACING*CHAR_SIZE_MULT);

  // Flags
  if (v & 0x0001) {
    Tft.drawString("1", x, y, CHAR_SIZE_MULT, WHITE);
  } else {
    Tft.fillRectangle(x, y, 1*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  }
  x += 2*CHAR_X_W*CHAR_SIZE_MULT;
  if (v & 0x0002) {
    Tft.drawString("2", x, y, CHAR_SIZE_MULT, WHITE);
  } else {
    Tft.fillRectangle(x, y, 1*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  }
}


void PrintIOPins(uint8_t l)
{
  bool t;
  uint16_t x, y;

  x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT;
  y = CHAR_Y_OFFSET + (l * LINE_SPACING*CHAR_SIZE_MULT);

  t = digitalRead(NIGHT_SNS) == HIGH;
  if (t != nightAsserted) {
    if (t) {
      Tft.drawString("N", x, y, CHAR_SIZE_MULT, WHITE);
    } else {
      Tft.fillRectangle(x, y, 1*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
    }
    nightAsserted = t;
  }

  x += 2*CHAR_X_W*CHAR_SIZE_MULT;
  t = digitalRead(ALERT_SNS) == LOW;
  if (t != alertAsserted) {
    if (t) {
      Tft.drawString("A", x, y, CHAR_SIZE_MULT, WHITE);
    } else {
      Tft.fillRectangle(x, y, 1*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
    }
    alertAsserted = t;
  }

  x += 2*CHAR_X_W*CHAR_SIZE_MULT;
  t = digitalRead(PWR_SNS) == HIGH;
  if (t != pwrEnabled) {
    if (t) {
      Tft.drawString("5", x, y, CHAR_SIZE_MULT, WHITE);
    } else {
      Tft.fillRectangle(x, y, 1*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
    }
    pwrEnabled = t;
  }
}


void PrintWdtResetDetected(bool en)
{
  uint16_t x, y;

  x = CHAR_X_W_OFFSET;
  y = CHAR_Y_OFFSET + (LB_SYS_WD_LINE * LINE_SPACING*CHAR_SIZE_MULT);

  if (en) {
    Tft.drawString("WDT Reset", x, y, CHAR_SIZE_MULT, WHITE);
  } else {
    Tft.fillRectangle(x, y, 9*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  }
}


void PrintWatchdogDetected(bool en)
{
  uint16_t x, y;

  x = CHAR_X_W_OFFSET;
  y = CHAR_Y_OFFSET + (LB_PWR_WD_LINE * LINE_SPACING*CHAR_SIZE_MULT);

  if (en) {
    Tft.drawString("WDT Power", x, y, CHAR_SIZE_MULT, WHITE);
  } else {
    Tft.fillRectangle(x, y, 9*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
  }
}


void PrintWatchdogCount(bool en)
{
  static uint16_t prevCount = 999;  // Bigger than 8-bits
  static uint32_t prevPwrOff = 99999;  // Bigger than 16-bits;
  uint8_t t;
  uint16_t t16, x, y;

  if (en) {
    // Update value
    if (Read8(SMB_ADDR_WD_TO, &t)) {
      if (t != prevCount) {
        prevCount = t;
        PrintNum16(LB_WD_COUNT_LINE, 0, prevCount);
      }
    }
    if (Read16(SMB_ADDR_WD_PWROFF, &t16)) {
      if ((uint32_t) t16 != prevPwrOff) {
        prevPwrOff = t16;
        PrintNum16(LB_WD_COUNT_LINE, 6, t16);
      }
    }
  } else {
    if (prevCount != 999) {
      // Delete text
      x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT;
      y = CHAR_Y_OFFSET + (LB_WD_COUNT_LINE * LINE_SPACING*CHAR_SIZE_MULT);
      prevCount = 999;
      Tft.fillRectangle(x, y, 6*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
    }
    if (prevPwrOff != 99999) {
      // Delete text
      x = CHAR_X_W_OFFSET + CHAR_DATA_OFF*CHAR_SIZE_MULT + 6*CHAR_X_W*CHAR_SIZE_MULT;
      y = CHAR_Y_OFFSET + (LB_WD_COUNT_LINE * LINE_SPACING*CHAR_SIZE_MULT);
      prevPwrOff = 99999;
      Tft.fillRectangle(x, y, 6*CHAR_X_W*CHAR_SIZE_MULT, CHAR_Y_W*CHAR_SIZE_MULT, BLACK);
    }
  }
}


void LogStatus()
{
 uint16_t t = data[SMB_INDEX_STATUS];

 Serial.print(F("Status: "));
 Serial.println(t, HEX);
 if (t & SMB_ST_SWD_DET_MASK) Serial.println(F("  SYS_WD"));
 if (t & SMB_ST_PWD_TRIG_MASK) Serial.println(F("  POWER_WD"));
 if (t & SMB_ST_BAD_BATT_MASK) Serial.println(F("  BAD_BATT"));
 if (t & SMB_ST_ET_MISS_MASK) Serial.println(F("  ET_MISSING"));
 if (t & SMB_ST_WD_RUN_MASK) Serial.println(F("  WD_RUNNING"));
 if (t & SMB_ST_PWR_EN_MASK) Serial.println(F("  PWR_EN"));
 if (t & SMB_ST_ALERT_MASK) Serial.println(F("  ALERT_ASSERTED"));
 if (t & SMB_ST_PCTRL_MASK) Serial.println(F("  PCTRL_ASSERTED"));
 if (t & SMB_ST_T_LIM_MASK) Serial.println(F("  T_LIM"));
 if (t & SMB_ST_NIGHT_MASK) Serial.println(F("  NIGHT"));
 Serial.print(F("  CHG ST: "));
 Serial.println(t&SMB_ST_CHG_ST_MASK);
}


void EvalCommand() {
  char c;
  
  if (Serial.available()) {
    c = Serial.read();
    
    switch (CmdState) {
      case CMD_ST_IDLE:
        if (ValidateCommand(c)) {
          CurCmd = c;
          CmdHasVal = false;
          CmdArg = 0;
          CmdVal = 0;
          CmdState = CMD_ST_CMD;
        }
        break;
        
      case CMD_ST_CMD:
        if (c == TERM_CHAR) {
          ProcessCommand();
          CmdState = CMD_ST_IDLE;
        } else {
          if ((c >= '0') && (c <= '9')) {
            CmdArg = (CmdArg*10) + (c-'0');
          } else if (c == '=') {
            CmdState = CMD_ST_VAL;
          } else {
            CmdState = CMD_ST_IDLE;
          }
        }
        break;
        
      case CMD_ST_VAL:
      if (c == TERM_CHAR) {
          ProcessCommand();
          CmdState = CMD_ST_IDLE;
        } else {
          if ((c >= '0') && (c <= '9')) {
            CmdVal = (CmdVal*10) + (c-'0');
            CmdHasVal = true;
          } else {
            CmdState = CMD_ST_IDLE;
          }
        }
        break;

      default:
        CmdState = CMD_ST_IDLE;
    }
  }
}


boolean ValidateCommand(char c) {
  switch (c) {
    case 'A':
    case 'C':
    case 'D':
    case 'H':
    case 'L':
    case 'P':
    case 'S':
    case 'T':
    case 'W':
      return true;
      break;
    
    default:
      return false;
  }
}


void ProcessCommand() {
  uint16_t t16;
  uint8_t t8;

  switch (CurCmd) {
    case 'A':
      if (CmdHasVal) {
        if (CmdVal == 0) {
          logMode = MODE_IDLE;
        } else {
          logMode = MODE_LCD;
        }
      }
      if (logMode == MODE_IDLE) {
        Serial.println(F("LCD Update disabled"));
      } else {
        Serial.println(F("LCD Update enabled"));
      }
      break;
      
    case 'C':
      PrintWdtResetDetected(false);
      PrintWatchdogDetected(false);
      break;
      
    case 'D':
      if (!CmdHasVal) {
        if (CmdArg < SMB_RO_NUM) {
          t16 = data[CmdArg];
          if (CmdArg == SMB_INDEX_ID) {
            Serial.print(t16 >> 12);
            Serial.print(F(" "));
            Serial.print((t16 >> 4) & 0x0F);
            Serial.print(F("."));
            Serial.println(t16 & 0x0F);
          } else if (CmdArg == SMB_INDEX_STATUS) {
            Serial.print(F("0x"));
            Serial.println(t16, HEX);
          } else if (CmdArg == SMB_INDEX_BUCK) {
            Serial.print(t16 >> 8);
            Serial.print(F(" "));
            Serial.println(t16 & 0xFF);
          } else {
            Serial.println((int16_t) t16);
          }
        }
      }
      break;
      
    case 'H':
      DisplayHelp();
      break;
    
    case 'L':
      if (CmdHasVal) {
        logCount = CmdVal;
        if (CmdVal == 0) {
          if (logMode != MODE_LCD) {
            logMode = MODE_LCD;
            SetupDisplayLcdLog();
          }
        } else {
          if (logTapTrigEn) {
            logMode = MODE_WAIT_TAP;
            SetupDisplayWaitTap();
          } else {
            logMode = MODE_SER;
            SetupDisplaySerLog();
          }
        }
      }
      Serial.print(F("Log Count = "));
      Serial.println(logCount);
      break;
      
    case 'P':
      if (CmdHasVal) {
        if ((CmdArg >= 0) && (CmdArg <= 3)) {
          if (!Write16(CmdArg*2 + SMB_ADDR_BULK_V, CmdVal)) {
            Serial.println(F("Write failed"));
          }
        }
      } else {
        if (Read16(CmdArg*2 + SMB_ADDR_BULK_V, &t16)) {
          Serial.println(t16);
        } else {
          Serial.println(F("Read failed"));
        }
      }
      break;

    case 'S':
      LogStatus();
      break;
      
    case 'T':
      if (CmdHasVal) {
        logTapTrigEn = (CmdVal > 0);
      }
      Serial.print(F("Tap Trig En = "));
      Serial.println(logTapTrigEn);
      break;
      
    case 'W':
      if (CmdHasVal) {
        if ((CmdArg == 0) || (CmdArg == 1)) {
          if (!Write8(CmdArg*2 + SMB_ADDR_WD_EN, CmdVal)) {
            Serial.println(F("Write failed"));
          }
        } else if (CmdArg == 2) {
          if (!Write16(SMB_ADDR_WD_PWROFF, CmdVal)) {
            Serial.println(F("Write failed"));
          }
        }
      } else {
        if ((CmdArg == 0) || (CmdArg == 1)) {
          if (Read8(CmdArg*2 + SMB_ADDR_WD_EN, &t8)) {
            Serial.println(t8);
          } else {
            Serial.println(F("Read failed"));
          }
        } else if (CmdArg == 2) {
          if (Read16(SMB_ADDR_WD_PWROFF, &t16)) {
            Serial.println(t16);
          } else {
            Serial.println(F("Read failed"));
          }
        }
      }
      break;
  }
}


void DisplayHelp()
{
  Serial.println(F("Command Interface:"));
  Serial.println(F("  Misc commands:"));
  Serial.println(F("   A        : Control LCD Update (0=disable, 1=enable)"));
  Serial.println(F("   C        : Clear Watchdog Detected indication"));
  Serial.println(F("   H        : Print this list"));
  Serial.println(F("   S        : Print analysis of most recent status"));
  Serial.println(F("  Data Memory (N=0-11)"));
  Serial.println(F("   D<N>     : Display data memory location"));
  Serial.println(F("  Serial Dump"));
  Serial.println(F("   L=<N>    : Setup to or start logging data to the serial port (N is number of samples,"));
  Serial.println(F("              N < 32768"));
  Serial.println(F("   T=<N>    : Configure tap-on-lcd trigger for logging (0 = command starts immediately,"));
  Serial.println(F("              1 = user must tap on lcd to start logging)"));
  Serial.println(F("  Parameter Memory (N=0-3):"));
  Serial.println(F("   P<N>     : Read parameter memory"));
  Serial.println(F("   P<N>=<V> : Write parameter memory"));
  Serial.println(F("  Watchdog Registers (N=0-2):"));
  Serial.println(F("   W<N>     : Read a watchdog register"));
  Serial.println(F("   W<N>=<V> : Write a watchdog register (W0=234 to enable)"));
}

